/* ====================================================================
 * OpenCMS Lucene Search Engine Integration
 * 
 * Copyright (C) 2002 Syskinetic Consulting Limited.
 * http://www.syskinetic.com
 * 
 * Copyright (C) 2002 AP Soft Ltd.
 * http:// www.apsoft.ch
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 ====================================================================
*/

import java.io.*;
import java.util.*;
import java.math.*;

import com.opencms.file.*;
import com.opencms.core.*;
import com.opencms.util.*;
import com.opencms.template.*;

import org.apache.lucene.analysis.*;
import org.apache.lucene.index.*;
import org.apache.lucene.document.*;
import org.apache.lucene.search.*;
import org.apache.lucene.queryParser.*;


public class Lucene extends CmsXmlTemplate implements com.opencms.workplace.I_CmsWpConstants {
	
	// resource types definitions
	static final int C_FILE_HTML = 1;
	static final int C_FILE_XML = 2;
	static final int C_SYSTEM_XML = 3;
	static final int C_SYSTEM_HTML = 4;
	static final int C_FILE_DOWNLOAD = 5;
	static final int C_FILE_IMAGE = 6;

	// index document fields
	static final String FIELD_URL = "url";
	static final String FIELD_MODIFIED = "modified";
	static final String FIELD_TITLE = "title";
	static final String FIELD_CONTENT = "content";
	static final String FIELD_SUMMARY = "summary";

	// number of results displayed per page
	static final int HITS_PER_PAGE = 5;
  
	// used to tokenize queries
	protected Analyzer analyzer = new StopAnalyzer();
  
	
	public Object createIndex(CmsObject cms, String tagcontent, A_CmsXmlContent doc, Object userObject) throws CmsException {

		String indexName = getIndexDir(doc);
		try {
			IndexWriter writer = new IndexWriter(indexName, new SimpleAnalyzer(), true);
			writer.mergeFactor = 20;
			writer.optimize();
			writer.close();
		}
		catch (Exception e) {
			System.out.println(" caught a " + e.getClass() + "\n with message: " + e.getMessage() + "\n in createIndex");
		}
		return "";
	}
  
	public Object indexContent(CmsObject cms, String tagcontent, A_CmsXmlContent doc, Object userObject) throws CmsException {

		String result = null;
		String fileName = tagcontent;
		String indexName = getIndexDir(doc);
		CmsXmlTemplateFile template=(CmsXmlTemplateFile)doc;

		result = addContent(cms, template, indexName, fileName);
		return result.getBytes();
	}

	public Object searchIndex(CmsObject cms, String tagcontent, A_CmsXmlContent doc, Object userObject) throws CmsException, IOException {

		int start =0;
		Query query = null;
		String indexName = getIndexDir(doc);
		String queryName=tagcontent;
		String servletPath = cms.getRequestContext().getRequest().getServletUrl();
		Hashtable parameters = (Hashtable) userObject;
		String queryString = (String) parameters.get(queryName);
		CmsXmlTemplateFile template=(CmsXmlTemplateFile)doc;
		StringBuffer result = new StringBuffer();
		Searcher searcher =	new IndexSearcher(getReader(indexName));
    
		try {
			String startString = (String) parameters.get("start");
			if((queryString==null) || (queryString=="")) 
				return "";
			template.setData("query", Entities.encode(queryString));
			template.setData("hits", "-1");

			try {
				query = QueryParser.parse(queryString, FIELD_CONTENT, analyzer);
			}
			catch (ParseException e) {
				return "";
			}
			catch (TokenMgrError e) {
				return "";
			} 

			Hits hits = searcher.search(query);
			try {
				start = (startString != null)?Integer.parseInt(startString)-1:0;
			}
			catch (Exception e) { start=0;}


			// ensure start parameter is within bounds
			start = (start < 1) ? 0:(start >= hits.length()) ? hits.length()-1:start;  
			int end = Math.min(hits.length(), start + HITS_PER_PAGE);

			for (int i = start; i < end; i++) {
				Document docmt = hits.doc(i);
				template.setData("url",servletPath+docmt.get("url"));
				template.setData("title",docmt.get("title"));
				template.setData("score",""+Math.round(100 * hits.score(i)));
				template.setData("path",docmt.get("path"));
				template.setData("count",""+(i+1));
				String summary = docmt.get("summary");
				if (summary==null) {summary="";}
				template.setData("summary",summary);
				result.append(template.getProcessedDataValue("result", this, userObject));
			}

			template.setData("start",""+(start+1));
			template.setData("queryName",queryName);
			template.setData("hits",""+hits.length());
			template.setData("end",""+end);

			return result.toString();
		}
		finally  {
			if (searcher != null) searcher.close(); // enabled only if uncached IndexReader
		}
	}

	public Object showPrevious(CmsObject cms, String tagcontent, A_CmsXmlContent doc, Object userObject) throws CmsException {
		CmsXmlTemplateFile template = (CmsXmlTemplateFile) doc;
		StringBuffer result = new StringBuffer();

		String startString = template.hasData("start")?template.getDataValue("start"):"";
		int start = (startString != "")?Integer.parseInt(startString):1;
		String queryName = template.hasData("queryName")?template.getDataValue("queryName"):"";
		String query = template.hasData("query")?template.getDataValue("query"):"";
		String hitsString = template.hasData("hits")?template.getDataValue("hits"):"";
		int hits = (hitsString != "")?Integer.parseInt(hitsString):0;

		if((start - HITS_PER_PAGE) >= 1) {
			template.setData("url", "?" + queryName + "=" + query + "&start=" + (start - HITS_PER_PAGE));
			result.append(template.getProcessedDataValue("previous", this, userObject));
		}
		return result.toString();
	}

	public Object showNext(CmsObject cms, String tagcontent, A_CmsXmlContent doc, Object userObject) throws CmsException {
		CmsXmlTemplateFile template = (CmsXmlTemplateFile) doc;
		StringBuffer result = new StringBuffer();

		String startString = template.hasData("start")?template.getDataValue("start"):"";
		int start = (startString != "")?Integer.parseInt(startString):1;
		String queryName = template.hasData("queryName")?template.getDataValue("queryName"):"";
		String query = template.hasData("query")?template.getDataValue("query"):"";
		String hitsString = template.hasData("hits")?template.getDataValue("hits"):"";
		int hits = (hitsString != "")?Integer.parseInt(hitsString):0;

		if(hits >= (start + HITS_PER_PAGE)) {
			template.setData("url","?" + queryName + "=" + query + "&start=" + (start + HITS_PER_PAGE));
			result.append(template.getProcessedDataValue("next", this, userObject));
		}
		return result.toString();
	}

	public Object showMessage(CmsObject cms, String tagcontent, A_CmsXmlContent doc, Object userObject) throws CmsException {
	
		CmsXmlTemplateFile template = (CmsXmlTemplateFile) doc;
		StringBuffer result = new StringBuffer();
		String hitsString = template.hasData("hits")?template.getDataValue("hits"):"";
		int hits = (hitsString != "")?Integer.parseInt(hitsString):0;
		String message = (hits > 0)?"message1":(hits < 0)?"error":"message0";
		if(!template.hasData("query"))
			template.setData("query","");

		result.append(template.getProcessedDataValue(message, this, userObject));
		return result.toString();
	}

	// get a property of the current folder
	public Object getPropertyCurrent(CmsObject cms, String tagcontent, A_CmsXmlContent doc, Object userObject) throws CmsException {

		String property= "";
		CmsXmlTemplateFile template= (CmsXmlTemplateFile) doc;
		String propertyname = tagcontent;
		if (!propertyname.equals("")) {
			String path = template.hasData("path")?template.getDataValue("path"):"";
			if((path==null) || (path==""))
				path=cms.getRequestContext().currentFolder().getAbsolutePath();

			property = cms.readProperty(path, propertyname);
			property = (property != null) ? property:"";
		}
		return (property.getBytes());
	}

	
	
	// private methods /////////////////////////////////////////////////////////
	
	private void addDocument(IndexWriter writer,
							String url,
							String modified,
							String title,
							String content,
							String summary) throws IOException {
		
		Document docmt = new Document();
		docmt.add(Field.Text(FIELD_URL, url));
		docmt.add(Field.Text(FIELD_MODIFIED, modified));
		docmt.add(Field.Text(FIELD_TITLE, title));
		docmt.add(Field.Text(FIELD_CONTENT, content));
		docmt.add(Field.Text(FIELD_SUMMARY, summary));
		writer.addDocument(docmt);
	}
	
	private void addDocument(IndexWriter writer,
							String url,
							String modified,
							String title,
							Reader reader,
							String summary) throws IOException {
		
		Document docmt = new Document();
		docmt.add(Field.Text(FIELD_URL, url));
		docmt.add(Field.Text(FIELD_MODIFIED, modified));
		docmt.add(Field.Text(FIELD_TITLE, title));
		docmt.add(Field.Text(FIELD_CONTENT, reader));
		docmt.add(Field.Text(FIELD_SUMMARY, summary));
		writer.addDocument(docmt);
	}
	
	private String displayIndexedFile(CmsXmlTemplateFile template, String path) throws CmsException {
		template.setData("item", "file: " + path);
		return template.getProcessedDataValue("addItem");
	}
	
	private String displayRemovedFile(CmsXmlTemplateFile template, String path) throws CmsException {
		template.setData("item", "file: " + path);
		return template.getProcessedDataValue("removeItem");
	}
	
	private String getIndexDir(A_CmsXmlContent doc) throws CmsException {
  	 
		CmsXmlTemplateFile template=(CmsXmlTemplateFile)doc;
		String index = (String) template.getDataValue("indexDirectory");
		return (index == null) ? "index":index;	
	}
	
	private LinkedList createDocumentList(IndexReader reader) throws IOException {
		int max = reader.numDocs();
		LinkedList list = new LinkedList();
		
		for (int i = 0; i < max; i++) {
			if (!reader.isDeleted(i)) {
				Document doc = reader.document(i);
				IndexDocument idoc = new IndexDocument(i, doc.get(FIELD_URL), doc.get(FIELD_MODIFIED));
				list.add(idoc);
			}
		}
		return list;
	}
	
	private String removeDocuments(CmsXmlTemplateFile template, IndexReader reader, LinkedList docList) throws IOException,CmsException {
		String result = "";
		int max = docList.size();
		// remove remaining docunments from index (deleted or modified)
		for (int i = 0; i < max; i++) {
			IndexDocument idoc = (IndexDocument) docList.get(i);
			reader.delete(idoc.id);
			System.out.println("Removing: " + idoc.url);
			result += displayRemovedFile(template, idoc.url);
		}
		return result;
	}
	
	private String addContent(CmsObject cms,CmsXmlTemplateFile template, String indexName, String root) throws CmsException {
		
		IndexReader reader = null;
		IndexWriter writer = null;
		
		try {
			ArrayList fileList = new ArrayList();
			// check index differences
			reader = IndexReader.open(indexName);
			LinkedList docList = createDocumentList(reader);
			makeIndexDiff(cms, docList, root, fileList);
			String result = removeDocuments(template, reader, docList);
			reader.close();
			// update index
			writer = new IndexWriter(indexName, new SimpleAnalyzer(), false);
			result += updateIndex(cms, template, writer, fileList.iterator());
			writer.optimize();
			writer.close();
			return result;
		}
		catch (Exception e) {
			try {
				if (reader != null) reader.close();
				if (writer != null) writer.close();
			} catch (IOException e1) {
			}
			throw new CmsException("Exception in addContent() " + e.getMessage());
		}
	}
	
	private String updateIndex(CmsObject cms,CmsXmlTemplateFile template,IndexWriter writer, Iterator fileList) throws Exception {
		
		StringBuffer result = new StringBuffer();
		String superfolder = com.opencms.boot.CmsBase.getWebBasePath();
		
		while (fileList.hasNext()) {
			IndexFile file = (IndexFile) fileList.next();
			
			switch(file.getType()) {
				case C_FILE_HTML:
					result.append(processHtml(cms, file.getPath(), writer, template,file.getModified()));
					break;

				case C_FILE_DOWNLOAD:
					String bodyFilename = superfolder + file.getPath().substring(1).replace('/',File.separatorChar);
					if (bodyFilename.lastIndexOf(".pdf") != -1)
						result.append(processPdf(cms, bodyFilename, file.getPath(), writer, template, file.getModified()));
					else if (bodyFilename.lastIndexOf(".doc") != -1)
						result.append(processDoc(cms, bodyFilename, file.getPath(), writer, template, file.getModified()));
					else if (bodyFilename.lastIndexOf(".rtf") != -1)
						result.append(processRtf(cms, bodyFilename, file.getPath(), writer, template, file.getModified()));
					else
						;//result.append(processUnknown(cms, bodyFilename, file.getPath(), writer, template, file.getModified()));
					break;

				default:
					continue;
			}
		}
		return result.toString();
	}
	
	private void makeIndexDiff(CmsObject cms, LinkedList docList, String root, Collection fileList) throws Exception {
		
		String path = null;
		CmsFolder folder = null;
		// get all subfolders of the root folder 
		Vector folders = cms.getSubFolders(root);
		Vector files = cms.getFilesInFolder(root);
		Enumeration enum = files.elements();
		
		while(enum.hasMoreElements()) {
			CmsFile file = (CmsFile) enum.nextElement();
			path = file.getAbsolutePath();
			String modified = Long.toString(file.getDateLastModified());
			int fileType = file.getType();
			
			// check if the resource is not marked as deleted
			if(file.getState() != C_STATE_DELETED && (file.getAccessFlags() & C_ACCESS_PUBLIC_READ) !=0) {
				if (isFileSupported(fileType, path)) {
					IndexFile indexFile = processFile(docList, path, modified, fileType);
					if (indexFile != null)
						fileList.add(indexFile);
				}
			}
		}
    
		enum = folders.elements();
		while(enum.hasMoreElements()) {
			folder = (CmsFolder)enum.nextElement();
			if((folder.getAccessFlags() & C_ACCESS_INTERNAL_READ) == 0 ) {
				path = folder.getAbsolutePath();
				makeIndexDiff(cms,docList,path,fileList);
			}
		}
	}
  
	private boolean isFileSupported(int type, String file) {
		switch(type) {
			case C_FILE_HTML:
				return true;
			case C_FILE_DOWNLOAD:
				if (file.lastIndexOf(".pdf") != -1 
					|| file.lastIndexOf(".doc") != -1
					|| file.lastIndexOf(".rtf") != -1)
					return true;
		}
		return false;
	}
	
	private int findDocument(LinkedList docList, String url) {
		int num = -1;
		int max = docList.size();
		
		for (int i = 0; i < max; i++) {
			IndexDocument idoc = (IndexDocument) docList.get(i);
			if (idoc.url.equals(url)) {
				num = i;
				break;
			}
		}
		return num;
	}

	private IndexFile processFile(LinkedList docList, String url, String modified, int type) {
				
		IndexFile indexFile = null;
		int docNum = findDocument(docList, url);

		if (docNum == -1) { 
			// new entry to be added to index
			indexFile = new IndexFile(url, modified, type);
			System.out.println("New File: " + url); //DEBUG
		}
		else {
			// compare last modified date
			IndexDocument idoc = (IndexDocument) docList.get(docNum);
			if (idoc.modified.equals(modified)) {
				// file is unchanged - remove it from deletion list
				docList.remove(docNum);
				System.out.println("Existing - not modified: " + url); //DEBUG
			}
			else {
				// file was modified must be updated
				indexFile = new IndexFile(url, modified, type);
				System.out.println("Existing - must be modified: " + url); //DEBUG
			}
		}
		return indexFile;
	}
	
	private String processHtml(CmsObject cms, String path, IndexWriter writer,
  						CmsXmlTemplateFile template, String modified) throws Exception {
		
		CmsXmlControlFile controlFile = new CmsXmlControlFile(cms, path);
		String bodyElementClassName = controlFile.getElementClass(C_BODY_ELEMENT);
		Object tempObj = CmsTemplateClassManager.getClassInstance(cms, bodyElementClassName);
		CmsXmlTemplate bodyElementClassObject = (CmsXmlTemplate) tempObj;
		String bodyFilename = C_CONTENTBODYPATH + path.substring(1);
		CmsXmlTemplateFile bodyTemplateFile = bodyElementClassObject.getOwnTemplateFile(cms, bodyFilename, C_BODY_ELEMENT, null, null);
		String content = HTMLParser.stripText(bodyTemplateFile.getTemplateContent(this, null, null));
		String summary = summary(content,200);

		String title = cms.readProperty(path,C_PROPERTY_TITLE);
		if (title == null)
			title = path;

		addDocument(writer, path, modified, title, content, summary);
		return displayIndexedFile(template, path);
	}
	
	private String processPdf(CmsObject cms, String file, String path, IndexWriter writer,
  						CmsXmlTemplateFile template, String modified) throws Exception {
		try {
			File f = new File(file);
			PdfExtractor pdf = new PdfExtractor();
			String content = pdf.extractText(file);
			String summary = summary(content,200);

			String title = cms.readProperty(path,C_PROPERTY_TITLE);
			if(title == null) 
				title = f.getName();

			addDocument(writer, path, modified, title, content, summary);
			return displayIndexedFile(template, path);
		}
		catch (IOException e) { 
			System.out.println("bad file not indexed :" + file);
			return "";
		}
	}

	private String processPdfBatch(CmsObject cms, String file, String path, IndexWriter writer,
  						CmsXmlTemplateFile template, String modified) throws Exception {
		try {
			File f = new File(file);
			String batchFile = template.getDataValue("batchFile");
			String outputDir = template.getDataValue("outputDir");

			BatchPdfExtractor pdf = new BatchPdfExtractor();
			String content = pdf.extractText(file, batchFile, outputDir);
			if (content == null)
				throw new IOException("");
			content = HTMLParser.stripText(content);
			String summary = summary(content,200);

			String title = cms.readProperty(path,C_PROPERTY_TITLE);
			if (title == null)
				title = f.getName();

			addDocument(writer, path, modified, title, content, summary);
			return displayIndexedFile(template, path);
		}
		catch (IOException e) { 
			System.out.println("bad file not indexed :" + file);
			return "";
		}
	}
  
	private String processDoc(CmsObject cms, String file, String path, IndexWriter writer,
  						CmsXmlTemplateFile template, String modified) throws Exception {
		System.out.println("Doc currently not supported: " + file);
		return "";
	}

	private String processRtf(CmsObject cms, String file, String path, IndexWriter writer,
  						CmsXmlTemplateFile template, String modified) throws Exception {
		try {
			javax.swing.text.rtf.RTFEditorKit rtf = new javax.swing.text.rtf.RTFEditorKit();
			javax.swing.text.Document doc = rtf.createDefaultDocument();

			File f = new File(file);
			FileInputStream is = new FileInputStream(f);
			rtf.read(is,doc,0);
			String content = doc.getText(0,doc.getLength()); 
			String summary = summary(content,200);
			is.close();

			String title = cms.readProperty(path,C_PROPERTY_TITLE);
			if (title == null) 
				title = f.getName();
			
			addDocument(writer, path, modified, title, content, summary);
			return displayIndexedFile(template, path);
		}
		catch (IOException e) { 
			System.out.println("bad file not indexed :" + file);
			return "";
		}
	}

	private String processUnknown(CmsObject cms, String file, String path, IndexWriter writer,
							CmsXmlTemplateFile template, String modified) throws Exception {
		try {
			File f = new File(file);
			FileInputStream is = new FileInputStream(f);
			Reader reader = new BufferedReader(new InputStreamReader(is));

			String title = cms.readProperty(path,C_PROPERTY_TITLE);
			if (title == null) 
				title = f.getName();

			addDocument(writer, path, modified, title, reader, "");
			return displayIndexedFile(template, path);
		}
		catch (IOException e) { 
			System.out.println("bad file not indexed :" + file);
			return "";
		}
	}

	private String summary(String content, int size) throws Exception {

		if(content.length() <= size)
			return content;
		// try to find end of word
		int space = content.indexOf(" ", size);
		if ((space == -1) || (space > size*1.1))
			return content.substring(0, size);
		
		return content.substring(0, space);
	}
	
	IndexReader getReader(String name) {
		IndexReader reader = null;
		try {
			reader = IndexReader.open(name);
		}
		catch (IOException e) {
			System.out.println(" caught a " + e.getClass() +
				"\n with message: " + e.getMessage() +
				"\n in getReader");
		}
		return reader;
	}
	
	// do not cache the search results page
	public CmsCacheDirectives getCacheDirectives(CmsObject cms, String templateFile,
			String elementName, Hashtable parameters, String templateSelector) {
		return new CmsCacheDirectives(false);
	}
}




